### ============================================================================
### WIP Generator
###
### This program is licenced under the GNU General Public License for more details.
### (c) copyright Pierre Lemaire, 2019.
###
### author: Pierre Lemaire
### version: 2019-03-29
### ____________________________________________________________________________

### PARAMETERS =================================================================

DIR="TR"   ## (non-empty) name of the directory to put all produced files.

## generation:
SEED=0                            ## seed to generate everything, if not NULL.
N.FACTOR=1                        ## factor for number of jobs per instance
JOBS <- 0:5                       ## list of instance generators to use
MACHINES <- 0:5                   ## list of instance generators to use
FORCE.GENERATE <- FALSE           ## force generation

## scheduling:
INSTS <- data.frame(              ## instances to use (per column)
  #k        1  2 3 4  5  6  7
  jobs  = c(0, 2,2,2, 3, 4, 4),
  machs = c(0, 0,1,2, 3, 4, 5)
)
FORCE.SCHEDULING <- FALSE         ## force scheduling

## measures:
FORCE.MEASURES <- FALSE           ## force computing measures
RUNNING.K      <- 0.1             ## prop. of points to use for running values

## graphs:
FORCE.GRAPHS <- FALSE             ## force drawing graphs
GRAPHS.BASENAME <- "%s@@@.jpg"    ## @@@ is replaced by graph-type, NULL for no file

## clearing:
FORCE.CLEARING <- FALSE           ## force computing clearing
CLEARING.STEPS     <- 1e2         ## number of steps to clear with Cap = mean(capa)
CLEARING.MAX.STEPS <- 1e3         ## max number of steps

## simulation:
FORCE.SIMULATION <- FALSE         ## force simulation
SIMULATION.INSTS <- 1:nrow(INSTS) ## indices of INSTS to use for simulation
SIMULATION.NEXP  <- 1e2           ## number of simulations
SIMULATION.N.FACTOR <- N.FACTOR   ## number of jobs in a simulation

### MISC. TOOLS ================================================================

remake <- function(dst, src=NULL, force=FALSE){
  ## Returns iff dst must be remade.
  ## PARAMS:
  ## - dst (filename): file to be remade
  ## - src ([vector of] filename[s]): sources required to build dst
  ## - force (boolean): force remake
  ## VALUE: TRUE if force is true or if dst is older than src.
  
  if (force) return (TRUE)
  if (is.null(dst)) return (TRUE)
  if (is.na(dst)) return (TRUE)
  if (!file.exists(dst)) return (TRUE)
  if (is.null(src)) return (FALSE) # dst exists and no src needed
  for(f in src){
    if (file.info(dst)$mtime < file.info(f)$mtime) return (TRUE)
  }
  return(FALSE)
}

clean <- function(instances=FALSE,schedules=FALSE,kpis=FALSE,clears=FALSE,
                  graphs=FALSE, simulations=FALSE, all=FALSE){
  ## Clean (remove) files corresponding to the given types.
  
  if (all || instances) system(sprintf("rm %s/*.dat",DIR))
  if (all || schedules) system(sprintf("rm %s/*.sch",DIR))
  if (all || kpis)      system(sprintf("rm %s/*_kpi.Rdata",DIR))
  if (all || clears)    system(sprintf("rm %s/*_clear.Rdata",DIR))
  if (all || graphs)    system(sprintf("rm %s/*.jpg",DIR))
  if (all || graphs)    system(sprintf("rm %s/*.pdf",DIR))
  if (all || simulations) system(sprintf("rm %s/*_e*",DIR))
}


### INSTANCE GENERATOR =========================================================

coerce.jobs <- function(jobs, wip.init){
  ## Coerces jobs to amount to a WIP level of wip.init at time 0, with no 
  ## further arrivals.
  ## PARAMS:
  ## - jobs (set of jobs): initial set of jobs, sorted by arrival!!!
  ## - wip.init (int): initial wip level
  ## VALUE: a new coerced set of jobs
  
  cs <- cumsum(jobs$workload)
  if (cs[length(cs)] <= wip.init) return(jobs)

  idx <- which.max(cs>=wip.init)
  if (length(idx) > 0) jobs <- jobs[1:idx,] # select first jobs up to wip.init
  jobs$arrival <- jobs$arrival - max(jobs$arrival) # translate before time 0
  return(jobs)
}

make.jobs <- function(jobsfile,g,n.factor=1, wip.init=NULL, force=FALSE){
  ## Creates and writes one set of jobs.
  ## PARAMS:
  ## - jobsfile (str): name of the file where jobs are written
  ## - g (int): idf of the generator to use
  ## - n.factor (float): fraction of the normal number of jobs
  ## - postfix (str): string to append to the normal name of file
  ## - wip.init (int): if not NULL, jobs are coerced to amount to a WIP level
  ##                   of wip.init at time 0, with no further arrivals.
  ## - force (bool): force generation.
  ## VALUE: none.
  
  write.jobs <- function(jobs, filename){
    ## sub: write a set of jobs
    write.table(jobs, filename, 
                row.names = FALSE, col.names = T, sep=";", quote = FALSE)
  }

  instance <- function(idf, priority=1, batch_family=1, recipe="", workload=1, arrival=0, duration=1){
    ## sub: construct a set of jobs
    return(data.frame(
      idf=idf,
      priority=priority,
      batch_family=batch_family,
      recipe=recipe,
      workload=workload,
      arrival=arrival,
      duration=duration
    ))
  }

  ## ____________________________________________
  if (!remake(jobsfile, force=force)) invisible() ## no need to remake
  
  message(sprintf("Generating Jobs to %s using %d:", jobsfile,g), 
          appendLF = FALSE)
  jobs <- NULL
  msg  <- NULL
  if (g==0){ ## _________________________________
    msg <- "Everything is fixed"
    N <- 1e3 * n.factor
    d.arrival <- 1/6  ### durée inter-arrivée exacte (10 minutes)
    d.process <- 4/6  ### durée de process exacte (40 minutes)
    WL <- d.process   ### expected workload 
    
    jobs <- instance(
      idf=1:N,
      priority=1,
      batch_family=1,
      recipe="",
      workload=WL,
      arrival=(1:N)*d.arrival,
      duration=d.process
    )
  }
  else if (g==1){ ## ____________________________
    msg <- "Everything is smooth"
    N <- 1e3 * n.factor
    d.arrival <- 1/6  ### durée inter-arrivée moyenne (10 minutes)
    d.process <- 4/6  ### durée de process moyenne (40 minutes)
    WL <- d.process   ### expected workload 
    
    jobs <- instance(
      idf=1:N,
      priority=1,
      batch_family=1,
      recipe="",
      workload=WL,
      arrival=cumsum(rexp(N, rate=1/d.arrival)),
      duration=rexp(N,rate=1/d.process)
    )
  }  
  else if (g==2){ ## ____________________________
    msg <- "Simple case 1 with 2 classes"
    N <- 1e3 * n.factor
    Ns <- c(round(N/3), N-round(N/3))
    d.arrival <- c()
    d.process <- c()
    WL <- c()
    
    d.arrival[1] <- 1  ### durée inter-arrivée exacte (60 minutes)
    d.process[1] <- 1  ### durée de process exacte (60 minutes)
    WL[1] <- d.process[1]   ### expected workload 
    d.arrival[2] <- 2/6  ### durée inter-arrivée moyenne (20 minutes)
    d.process[2] <- 4/6  ### durée de process moyenne (40 minutes)
    WL[2] <- d.process[2]   ### expected workload 
    
    jobs <- instance(
      idf=1:N,
      priority=rep(1:2,Ns),
      batch_family=1,
      recipe="",
      workload=rep(WL,Ns),
      arrival=c((1:Ns[1]-1)*d.arrival[1], 
                cumsum(rexp(1:Ns[2], rate=1/d.arrival[2]))
      ),
      duration=c(rep(d.process[1],Ns[1]),
                 rexp(Ns[2],rate=1/d.process))
    )
  }
  else if (g==3){ ## ____________________________
    msg <- "Mix of 3 types, with batches and recipes"
    N <- 1e3 * n.factor
    
    prios.f <- c(function() sample(c(5,10,15), size=1),
                 function() sample(c(1,5,5,5,7,7,7,10,20), size=1),
                 function() sample(c(2,17), size=1)
    )
    duras.f <- c(function() 0.02+abs(rnorm(1, 1/2, .1)),
                 function() 0.02+abs(rnorm(1, 1/4, 0.01)),
                 function() 0.02+rexp(1, 1)
    )
    types <- sample(1:3, size = N, replace = TRUE)
    prios <- sapply(types, function(t) prios.f[[t]]())
    recipes=c("","",1)[types]
    workloads=c(1/2, 1/4, 1)[types]
    durations <- sapply(types, function(t) duras.f[[t]]())
      
    jobs <- instance(
      idf=1:N,
      priority=prios,
      batch_family=types,
      recipe=recipes,
      workload=workloads,
      arrival=cumsum(rexp(N, rate=4)),
      duration=durations
    )
  }
  else if (g==4){ ## ____________________________
    msg <- "6 types with batches, waves of WIP"
    N <- 2e3 * n.factor

    d.arrival <- 1/60 ### durée inter-arrivée moyenne (1 minute)
    d.process <- 1/10 ### durée de process moyenne (6 minutes)
    WL <- d.process   ### expected workload 
    
    pack.size=100
    WL.of.pack=WL*pack.size/4 ## expected to be done on 1 machine with batch 4
    pack.arrival=0:(N/pack.size)*WL.of.pack
    arrivals <- cumsum(rexp(1:N, rate=1/d.arrival)) + sort(rep(pack.arrival,pack.size))[1:N]
    jobs <- instance(
      idf=1:N,
      priority = 1,
      batch_family = sample(1:6, size=N, replace=T),
      workload=WL,
      arrival=arrivals,
      duration=rexp(N,rate=1/d.process)
    )
  }
  else if (g==5){ ## ____________________________
    msg <- "Mix of 3 types, with batches and recipes"
    N <- 2e3 * n.factor
    
    prios.f <- c(function() sample(c(5,10,15), size=1),
                 function() sample(c(1,5,5,5,7,7,7,10,20), size=1),
                 function() sample(c(4,17), size=1)
    )
    duras.f <- c(function() 0.03+abs(rnorm(1, 1/3, .1)),
                 function() 0.02+abs(rnorm(1, 1/4, 0.01)),
                 function() 0.01+rexp(1, 2)
    )
    types <- sample(1:3, size = N, replace = TRUE)
    prios <- sapply(types, function(t) prios.f[[t]]())
    recipes=types
    workloads=c(1/2, 1/4, 1)[types]
    durations <- sapply(types, function(t) duras.f[[t]]())
    
    jobs <- instance(
      idf=1:N,
      priority=prios,
      batch_family=types,
      recipe=recipes,
      workload=workloads,
      arrival=cumsum(rexp(N, rate=24)),
      duration=durations
    )
  }
  else if (g==6){ ## ____________________________
    msg <- "TODO"
    N <- 1e3 * n.factor
    jobs <- NULL
  }
  else{ ## ______________________________________
    msg <- "ERR: unknown instance!"
  }
  
  message(sprintf(" %s", msg))
  if (!is.null(jobs)){
    jobs <- jobs[order(jobs$arrival),]
    if (!is.null(wip.init)) jobs <- coerce.jobs(jobs, wip.init)
    write.jobs(jobs, jobsfile)
  }
}

make.machines <- function(machinesfile, g, force=FALSE){
  ## Creates and writes one set of machines.
  ## PARAMS:
  ## - machinesfile (str): name of the file where machines are written
  ## - g (int): idf of the generator to use
  ## - force (bool): force generation.
  ## VALUE: none.

  write.machines <- function(machines,policy,filename){
    ## sub: write a set of machines
    
    write.table(machines, filename, 
                row.names = FALSE, col.names = T, sep=";", quote = FALSE)
    cat(sprintf("\n@POLICY;%s", policy), file=filename, append = T)
  }
  
  instance <- function(idf, speed=1, batch_size=1, setup="lambda: 0", recipes="[]"){
    ## sub: construct a set of machines
    
    return(data.frame(
      idf=idf,
      speed=speed,
      batch_size=batch_size,
      setup=setup,
      recipes=recipes
    ))
  }
  
  ## ____________________________________________
  if (!remake(machinesfile, force=force)) invisible() ## no need to remake
  
  message(sprintf("Generating Machines to %s using %d", machinesfile,g), 
          appendLF = FALSE)
  machines <- NULL
  policy <- NULL
  msg  <- NULL
  if (g==0){ ## _________________________________
    msg <- "One simple machine, FIFO policy"
    m <- 1
    policy <- "FIFO"
    machines <- instance(
      idf=1:m,
      speed=1,
      batch_size=1
    )
  }
  else if (g==1){ ## ____________________________
    msg <- "Two simple machines, FIFO policy"
    m <- 2
    policy <- "FIFO"
    machines <- instance(
      idf=1:m,
      speed=1,
      batch_size=1
    )
  }
  else if (g==2){ ## ____________________________
    msg <- "Two simple machines, different speed, WAIT-PRIO policy"
    m <- 2
    policy <- "WAIT-PRIO"
    machines <- instance(
      idf=1:m,
      speed=1:2,
      batch_size=1
    )
  }
  else if (g==3){ ## ____________________________
    msg <- "Three machines, 1 with recipe, different speeds, different batches, WAIT-PRIO policy"
    policy <- "WAIT-PRIO"
    machines <- instance(
      idf=1:3,
      speed=c(2,1,1),
      batch_size=c(1,2,3),
      recipes=c("[]", "[]", "[1]")
    )
  }
  else if (g==4){ ## ____________________________
    msg <- "Four machines, no recipe, no batch, FIFO policy"
    policy <- "FIFO"
    machines <- instance(
      idf=1:4,
      batch_size = 1
    )
  }
  else if (g==5){ ## ____________________________
    msg <- "One machine, no recipe, 4-batches, FIFO policy"
    policy <- "FIFO"
    machines <- instance(
      idf=1,
      batch_size = 4
    )
  }
  else if (g==6){ ## ____________________________
    msg <- "Five machines, with recipes, different speeds, different batches, WAIT-PRIO policy"
    policy <- "WAIT-PRIO"
    machines <- instance(
      idf=1:5,
      speed=c(1,1,2,2,3),
      batch_size=c(1,2,3,2,1),
      recipes=c("[1,2,3]", "[1,2]", "[2,1]", "[1,3]", "[1,2,3]")
    )
  }
  else{ ## ______________________________________
    msg <- "ERR: unknown instance!"
  }
  
  message(sprintf(" %s", msg))
  if (!is.null(machines)) write.machines(machines,policy,machinesfile)
}

### COMPUTATIONS ===============================================================

schedule <- function(schfile, jobsfile, machinesfile, nocwip=FALSE, force=FALSE, quiet=FALSE){
  ## Schedule jobs on machines, producing schfile as results.
  ## PARAMS:
  ## - schfile (str): name of the file where to write scheduling
  ## - jobsfile (str): name of the file from which jobs are read
  ## - machinesfile(str): name of the file from which machines are read
  ## - force (bool): force scheduling
  ## VALUE: none.
  
  if (!remake(schfile, c(jobsfile,machinesfile), force=force)) invisible()

  message(sprintf("Scheduling %s on %s to %s...", 
                  jobsfile, machinesfile, schfile))
  system(sprintf("./scheduler.py %s %s %s >%s %s", 
                 jobsfile, machinesfile, 
                 ifelse(nocwip, "--nocwip true", ""),
                 schfile,
                 ifelse(quiet, "2>/dev/null", "")))
}

mean.effective.capacity <- function(sch){
  ## Returns mean effective capacity of this schedule.
  return(sum(sch$WCap * sch$CWIP, na.rm = T)/sum(sch$CWIP, na.rm=T))
}

measures <- function(kpifile, schfile, sch=NULL, force=FALSE, running.k=RUNNING.K){
  ## Computes, writes and returns various measures on the schedule.
  ## PARAMS:
  ## - kpisfile (str): name of the file where to write KPI
  ## - schfile (str): name of the file from which schedule is read
  ## - sch (schedule): schedule, if already known
  ## - force (bool): force measures
  ## - running.k (float): proportion of values to keep for running-sd/quantiles
  ## VALUE: list of KPI...
  
  if (!remake(kpifile, schfile, force=force)){
    message(sprintf("Loading KPI for %s from %s...", schfile, kpifile))
    load(kpifile)
    invisible(kpi)
  }
  if (is.null(sch)) sch <- read.table(schfile, sep=";", header = T, na.strings = "None")

  message(sprintf("Computing KPI %s on %s...", kpifile, schfile))
  
  ## computes kpi:
  
  # wip:
  wip.max   <- max(sch$CWIP, na.rm = T)                # max level of WIP (as C-WIP)
  wip.limit <- quantile(sch$CWIP, probs=0.95, na.rm=T) # above, CWIP is not reliable
  
  # effective capacity:
  wcap.range <- range(sch$WCap, na.rm = T)   # range
  wcap.mean  <- mean.effective.capacity(sch) # (weighted) mean
  wcap.reg  <- loess(WCap ~ CWIP, data=sch)  # loess regression

  x <- sort(sch$CWIP)
  y <- sch$WCap[order(sch$CWIP)] 
  k <- max(length(y)*running.k, 2)
  wcap.rsd  <- caTools::runsd(y,k=k)                  # running SD
  wcap.rq95 <- caTools::runquantile(y,k=k,probs=0.95) # running 95%
  wcap.rq05 <- caTools::runquantile(y,k=k,probs=0.05) # running  5%
  wcap.reg95 <- loess(wcap.rq95 ~ x)                  # loess regression on 95%
  wcap.reg05 <- loess(wcap.rq05 ~ x)                  # loess regression on  5%
  
  ## pack, save and returns:
  kpi <- list(
    wip.max = wip.max,
    wip.limit = wip.limit,
    wcap.range = wcap.range,
    wcap.mean = wcap.mean,
    wcap.reg = wcap.reg,
    wcap.reg95 = wcap.reg95,
    wcap.reg05 = wcap.reg05,
    wcap.rsd = wcap.rsd,
    wcap.rq95 = wcap.rq95,
    wcap.rq05 = wcap.rq05
  )
  save(kpi, file=kpifile)
  invisible(kpi)
}

clearing <- function(clearfile, schfile, sch=NULL, kpifile, kpi=NULL, force=FALSE, 
                     steps=CLEARING.STEPS,max.steps=CLEARING.MAX.STEPS){
  ## Computes, writes and returns various clearing function.
  ## PARAMS:
  ## - clearfile (str): name of the file where to write Clearing process.
  ## - schfile (str): name of the file from which schedule is read
  ## - sch (schedule): schedule, if already known
  ## - kpifile (str): name of the file from which measures are read
  ## - kpi (measures): measures, if already known
  ## - force (bool): force measures
  ## - steps (int): number of steps to use
  ## VALUE: list of Clearing processes...
  
  if (!remake(clearfile, c(schfile, kpifile), force=force)){
    message(sprintf("Loading Clearing for %s from %s...", schfile, clearfile))
    load(clearfile)
    invisible(clear)
  }
  if (is.null(sch)) sch <- read.table(schfile, sep=";", header = T, na.strings = "None")
  if (is.null(kpi)) load(kpifile)
  
  message(sprintf("Computing Clearing %s on %s...", clearfile, schfile))
  
  ## computes clearing:
  wip0 <- kpi$wip.limit        # max level of wip
  t.mean = wip0 /kpi$wcap.mean # time to no-wip if Capacity = WCmean [constant]
  h <- t.mean/steps            # step size

  wip <- c(wip0,wip0,wip0)     # mean, 95%, 5%
  k <- 1
  cur <- wip
  while(k < max.steps){
    k <- k +1
    if (all(cur <= 0 | is.na(cur)))  break     # stop if done
    cap   <- c(predict(kpi$wcap.reg,   cur[1]),
               predict(kpi$wcap.reg95, cur[2]),
               predict(kpi$wcap.reg05, cur[3]))
    nex <- cur - cap*h
    wip <- cbind(wip,nex) ; cur <- nex
  }
  ts <- ((1:ncol(wip))-1) * h
  colnames(wip) <- NULL
  time.to.clear.reg   = ts[which(wip[1,]<=0)]
  time.to.clear.reg95 = ts[which(wip[2,]<=0)]
  time.to.clear.reg05 = ts[which(wip[3,]<=0)]
  
  if (length(time.to.clear.reg)==0)   time.to.clear.reg   = ts[which.min(wip[1,])]
  if (length(time.to.clear.reg95)==0) time.to.clear.reg95 = ts[which.min(wip[2,])]
  if (length(time.to.clear.reg05)==0) time.to.clear.reg05 = ts[which.min(wip[3,])]
  
  ## pack, save and returns:
  clear <- list(
    wip0 = wip0,
    instants = ts,
    clearing.reg = wip[1,],
    clearing.reg95 = wip[2,],
    clearing.reg05 = wip[3,],
    time.to.clear.mean = t.mean,
    time.to.clear.reg  = ifelse(length(time.to.clear.reg)   > 0, time.to.clear.reg[1],   NA),
    time.to.clear.reg95= ifelse(length(time.to.clear.reg95) > 0, time.to.clear.reg95[1], NA),
    time.to.clear.reg05= ifelse(length(time.to.clear.reg05) > 0, time.to.clear.reg05[1], NA)
  )
  save(clear, file=clearfile)
  invisible(clear)
  
}

### GRAPHS =====================================================================

start.graph <- function(filename){
  ## Start a graph, detecting the type.
  
  if (is.null(filename)) return()
  if (is.na(filename)) return()
  
  ext <- sub(".*\\.", "", filename)
  if      (ext == "pdf") pdf(filename)
  else if (ext == "jpg") jpeg(filename, width=800, height=800)
  else stop("Unkown graph type.")
}

end.graph <- function(filename){
  ## End a graph.
  if (!is.null(filename) && !is.na(filename)) dev.off()
}

graph.wip_vs_time <- function(graphfile, tag, schfile, sch=NULL, force=FALSE){
  ## Produces graphs of...
  ## PARAMS:
  ## - graphfile (str): name of the file where to write the graph. If NULL, produced on the current dev.
  ## - tag (str): tag to add to graph title.
  ## - schfile (str): name of the file from which schedule is read
  ## - sch (schedule): schedule, if already known
  ## - force (bool): force drawing graph
  ## VALUE: none.
  
  if (!remake(graphfile, schfile, force=force)) invisible()
  if (is.null(sch)) sch <- read.table(schfile, sep=";", header = T, na.strings = "None")
  
  message(sprintf("Graph WIP vs TIME of %s...", schfile))
  start.graph(graphfile)

  times <- unique(sort(c(sch$arrival,sch$start)))
  wip <- sapply(times, function(t) sum(sch$workload[sch$arrival<=t & sch$start>=t]))
  plot(wip ~ times, 
       t="l", lwd=2, col="purple",
       main=sprintf("WIP vs Time%s", tag),
       xlab="Time (hours)",
       ylab="WIP (workload)"
  )
  
  end.graph(graphfile)
}

graph.wc_vs_cwip <- function(graphfile, tag, schfile, sch=NULL, kpifile, kpi=NULL, force=FALSE,
                             annotations=TRUE){
  ## Produces graphs of...
  ## PARAMS:
  ## - graphname (str): name of the file where to write the graph. If NULL, produced on the current dev.
  ## - tag (str): tag to add to graph title.
  ## - schfile (str): name of the file from which schedule is read
  ## - sch (schedule): schedule, if already known
  ## - kpisfile (str): name of the file where to write KPI
  ## - kpi (measures): measures, if already known
  ## - force (bool): force drawing graph
  ## - annotations (bool): displays annotations (title, numbers...)
  ## VALUE: none.
  
  if (!remake(graphfile, c(schfile,kpifile), force=force)) invisible()
  if (is.null(sch)) sch <- read.table(schfile, sep=";", header = T, na.strings = "None")
  if (is.null(kpi)) load(kpifile)
  
  message(sprintf("Graph WC vs CWIP of %s...", schfile))
  start.graph(graphfile)
  plot(WCap~CWIP, data=sch, 
       col="black", pch=16, cex=.6,
       ylim=kpi$wcap.range + c(-.1, .1),
       main=ifelse(annotations,sprintf("Effective Capacity%s", tag),""),
       xlab="CWIP (workload)", ylab="Effective Capacity (workload/hours)")
  if (annotations) abline(h=kpi$wcap.mean, col="orange", lwd=2, lty=2)
  
  xs <- seq(0,kpi$wip.max, length.out=100)
  p.mean <- predict(kpi$wcap.reg,  xs)
  p.95   <- predict(kpi$wcap.reg95,xs)
  p.05   <- predict(kpi$wcap.reg05,xs)
  matlines(xs, cbind(p.mean, p.95, p.05), 
           col=c("blue",  "red",      "green"),
           lwd=c(3,       2,           2),
           lty=c(1,       1,           1)
  )

  end.graph(graphfile)
}

graph.hist_wc <- function(graphfile, tag, schfile, sch=NULL, kpifile, kpi=NULL, force=FALSE){
  ## Produces graphs of...
  ## PARAMS:
  ## - graphname (str): name of the file where to write the graph. If NULL, produced on the current dev.
  ## - tag (str): tag to add to graph title.
  ## - schfile (str): name of the file from which schedule is read
  ## - sch (schedule): schedule, if already known
  ## - kpisfile (str): name of the file where to write KPI
  ## - kpi (measures): measures, if already known
  ## - force (bool): force drawing graph
  ## VALUE: none.
  
  if (!remake(graphfile, c(schfile,kpifile), force=force)) invisible()
  if (is.null(sch)) sch <- read.table(schfile, sep=";", header = T, na.strings = "None")
  if (is.null(kpi)) load(kpifile)
  
  message(sprintf("Graph hist(WC) of %s...", schfile))
  start.graph(graphfile)
  
  hist(sch$WCap, 
       xlim=kpi$wcap.range + c(-.1, .1),
       main=sprintf("Effective capacities%s", tag), 
       xlab="Effective Capacity",
       probability = T)
  lines(density(sch$WCap,na.rm = T), col="gray")
  abline(v=kpi$wcap.mean, col="orange", lwd=2, lty=1)
  mtext(bquote(bar(C^E)==.(round(kpi$wcap.mean,2))), 3,line=-2, adj=1)
  
  end.graph(graphfile)
}

graph.clearing_wip <- function(graphfile, tag, schfile, sch=NULL, kpifile, kpi=NULL, clearfile, clear=NULL, force=FALSE, new.plot=TRUE, wip.init=NULL, annotations=TRUE){
  ## Produces graphs of...
  ## PARAMS:
  ## - graphname (str): name of the file where to write the graph. If NULL, produced on the current dev.
  ## - tag (str): tag to add to graph title.
  ## - schfile (str): name of the file from which schedule is read
  ## - sch (schedule): schedule, if already known
  ## - kpisfile (str): name of the file where to write KPI
  ## - kpi (measures): measures, if already known
  ## - clearsfile (str): name of the file where to write Clearing process
  ## - clear (measures): clearing process, if already known
  ## - force (bool): force drawing graph
  ## - new.plot (bool): force a new plot
  ## - wip.init (float): limit the graph to that amount of WIP
  ## - annotations (bool): displays annotations (title, numbers...)
  ## VALUE: none.
  
  if (!remake(graphfile, c(schfile,kpifile,clearfile), force=force)) invisible()
  if (is.null(sch)) sch <- read.table(schfile, sep=";", header = T, na.strings = "None")
  if (is.null(kpi)) load(kpifile)
  if (is.null(clear)) load(clearfile)
  
  message(sprintf("Graph Clearing of %s...", schfile))
  start.graph(graphfile)
  
  if (is.null(wip.init)) wip.init <- clear$clearing.reg[1]
  time.at.wip.init <- 0
  if (wip.init < clear$clearing.reg[1]){
    time.at.wip.init <- clear$instants[which.max(clear$clearing.reg < wip.init)]
  }    

  saved.par <- par(no.readonly = T)
  if (new.plot){
    plot(NA, 
         xlim=c(time.at.wip.init, clear$instants[length(clear$instants)]), 
         ylim=c(0,wip.init),
         main=ifelse(annotations, sprintf("Clearing of WIP%s", tag),""),
         xlab="Time (hours)",
         ylab="WIP (workload)"
    )
  }
  
  # clear with mean WC:  
  par(col="black", cex=.7)
  t = clear$time.to.clear.mean
  abline(b=-kpi$wcap.mean,a=clear$wip0, lwd=2, lty=2)
  if (annotations){ abline(v=t, lwd=1, lty=4) ; if (!is.na(t)) mtext(sprintf("%.2f",t),at=t,side=3) }
  
  # clear with WC(wip):  
  par(col="blue")
  t = clear$time.to.clear.reg
  lines(clear$clearing.reg ~ clear$instants, lwd=2, lty=2)
  if (annotations){ abline(v=t, lwd=1, lty=4) ; if (!is.na(t)) mtext(sprintf("%.2f",t),at=t,side=1) }
  
  # clear with mean WC(wip95):  
  par(col="red")
  t = clear$time.to.clear.reg95
  lines(clear$clearing.reg95 ~ clear$instants, lwd=2, lty=2)
  if (annotations){ abline(v=t, lwd=1, lty=4) ; if (!is.na(t)) mtext(sprintf("%.2f",t),at=t,side=1) }
  
  # clear with mean WC(wip05):  
  par(col="green")
  t = clear$time.to.clear.reg05
  lines(clear$clearing.reg05 ~ clear$instants, lwd=2, lty=2)
  if (annotations){ abline(v=t, lwd=1, lty=4) ; if (!is.na(t)) mtext(sprintf("%.2f",t),at=t,side=1) }
  
  # restore graph:
  par(col=saved.par$col,cex=saved.par$cex)

  end.graph(graphfile)
}

graph.simulation <- function(graphfile, tag, schfile, sch=NULL, kpifile, kpi=NULL, clearfile, clear=NULL,
                             force=FALSE,
                             wip.init.factor=0.75,
                             jobs.generator, machfile, n.exp=1e3, annotations=TRUE){
  ## Produces graphs of...
  ## PARAMS:
  ## - graphname (str): name of the file where to write the graph. If NULL, produced on the current dev.
  ## - tag (str): tag to add to graph title.
  ## - schfile (str): name of the file from which schedule is read
  ## - sch (schedule): schedule, if already known
  ## - kpisfile (str): name of the file where to write KPI
  ## - kpi (measures): measures, if already known
  ## - force (bool): force drawing graph
  ## - wip.init.factor (float): fraction of kpi$wip.limit to use as initial level of WIP for the simulations
  ## - jobs.generator (int): id of the generator to use
  ## - machfiles (str): name of the file from which machines are read
  ## - n.exp (int): number of simulations
  ## - annotations (bool): displays annotations (title, numbers...)
  ## VALUE: none.
  
  if (!remake(graphfile, c(schfile,kpifile,clearfile), force=force)) invisible()
  if (is.null(sch)) sch <- read.table(schfile, sep=";", header = T, na.strings = "None")
  if (is.null(kpi)) load(kpifile)
  if (is.null(clear)) load(clearfile)
  
  message(sprintf("Graph Simulation of %s...", schfile))
  start.graph(graphfile)
  
  #init graph:
  wip.init <- wip.init.factor*kpi$wip.limit
  time.at.wip.init <- clear$instants[which.max(clear$clearing.reg < wip.init)]
  suppressMessages(
    graph.clearing_wip(NULL,tag,schfile,sch, kpifile,kpi, clearfile,clear, force=T, wip.init=wip.init,
                       annotations=annotations)
  )
  #abline(v=time.at.wip.init, col="lightgray", lty=6)
  
  #add simulations:
  cmax <- c()
  mfile <- machfile
  for(e in 1:n.exp){
    message(sprintf("... %d", e), appendLF = F)
    jfile <- sprintf("%s/_e%03d.dat", DIR, e)
    sfile <- sprintf("%s/_e%03d.sch", DIR, e)
    
    # compute schedule:  
    
    suppressMessages(make.jobs(jfile,jobs.generator, SIMULATION.N.FACTOR, wip.init=wip.init))
    suppressMessages(schedule(sfile, jfile, mfile, nocwip=TRUE, force=T, quiet=T))
    
    # draw it:
    s <- read.table(sfile, sep=";", header = T, na.strings = "None")
    times <- unique(sort(c(0,s$start)))
    wip <- sapply(times, function(t) sum(s$workload[s$arrival <=t & s$start >=t]))
    lines(wip ~ I(times + time.at.wip.init), lwd=1, lty=1, col="lightgray")
    cmax <- c(cmax, max(times)+time.at.wip.init)
    #clean:
    system(sprintf("rm %s %s", jfile, sfile))
  }
  
  #redraw main graph:
  suppressMessages(graph.clearing_wip(NULL,"",schfile,sch, kpifile,kpi, clearfile,clear, force=T, new.plot = F,
                                      annotations=annotations))
  
  #add effective time.to.clearing:
  time.to.clearing.exp <- mean(cmax)
  if (annotations){
    abline(v=time.to.clearing.exp, col="gray", lty=4, lwd=2)
    mtext(sprintf("%.2f",time.to.clearing.exp),
          at=time.to.clearing.exp,side=1,line=1, col="gray")
  }  
  message()#EOL  
  end.graph(graphfile)
}
