#!/usr/bin/python3 -W once
# coding: utf8
# =======================================================================
'''Simple WIP simulator to evaluate CWIP and related stuff.

This program is licenced under the GNU General Public License for more details.
(c) copyright Pierre Lemaire, 2019.

@author: Pierre Lemaire
@version: 2019-03-28

'''

# APIs python:
import sys,logging,argparse
import fileinput,math

# ============================================================================
# Class for a Job
# ============================================================================
class Job:
    def __init__(self,idf,priority=1,batch=None,recipe=None,workload=1,arrival=0,duration=1):
        # init values:
        self.idf = idf           ## ID 
        self.priority = priority ## priority value (p1 < p2 <=> p1 is more important)
        self.batch_family = batch## only jobs of same batch_family can be batched together
        self.recipe = recipe     ## recipe (machine must know it)
        self.workload = workload ## a-priori workload (= expected duration)

        # operational values:
        self.arrival = arrival   ## effective arrival time
        self.duration = duration ## effective duration

        # scheduling values:
        self.start = None        ## effective start-time
        self.completion = None   ## effective completion-time
        self.machine = None      ## machine that process this job

        # CWIP values:
        self.CWIP = None           ## concurrent WIP (seen while waiting)
        self.WCap = None           ## witnessed capacity (seen while waiting)
        self.wip_at_arrival = None ## concurrent WIP (limited to what was present at arrival)

    ## -----------------------------------------------------------------------------
    def EWT(self):
        '''effective waiting time.'''
        if self.start is not None:
            return (self.start - self.arrival)
        else:
            return None

    ## -----------------------------------------------------------------------------
    def scheduleAt(self, t, completion=None, machine=None):
        '''set start-time at time t (and, optionally completion-time).'''
        self.start = t
        if completion is None:
            self.completion = self.start + self.duration
        else:
            self.completion = completion
        self.machine = machine
        
    ## -----------------------------------------------------------------------------
    @staticmethod
    def strfields():
        return("idf;priority;batch_family;recipe;workload;arrival;duration;start;completion;machine;EWT;CWIP;WCap;WipAtArrival")
    
    def __str__(self):
        return("%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s" \
        % (self.idf,self.priority,self.batch_family,self.recipe,self.workload,
           self.arrival,self.duration,
           self.start,self.completion,self.machine,self.EWT(),
           self.CWIP,self.WCap,self.wip_at_arrival))

    ## -----------------------------------------------------------------------------
    @staticmethod
    def read(filename):
        ''' Read jobs from file as a ;-separated file.'''
        log.info('Reading jobs from %s...' % filename)
        jobs = []
        for line in fileinput.input(filename):
            try:
                line = line.strip()
                if line == "": continue ## skip blank lines
                if line[0] == "#": continue ## skip comments
                    
                (idf,priority,batch_family,recipe,workload,arrival,duration) \
                  = line.split(";")
                if idf == 'idf': continue ## skip title line
                recipe = None if recipe == "" else int(recipe)
                    
                j = Job(idf, int(priority), int(batch_family), recipe,
                        float(workload), float(arrival), float(duration))
                jobs.append(j)
            except ValueError as e:
                log.error("%s:%d: %s" % (filename, fileinput.lineno(),e))

        log.info('... %d jobs' % len(jobs))
        return(jobs)

    ## -----------------------------------------------------------------------------
    @staticmethod
    def write(jobs):
        ''' Write jobs to stdout as a ;-separated file.'''
        log.info('Writing results to stdout...')
        print(Job.strfields())
        for job in jobs: print(job)

    ## -----------------------------------------------------------------------------
    @staticmethod
    def cwip(jobs):
        ''' Compute CWIP for a set of jobs.'''
        log.info('Computing CWIP...')
        log_count_total = len(jobs)
        log_count = 0

        jobs.sort(key=lambda j: j.arrival)
        for done,p in enumerate(jobs):
            if log_count == 0:
                log.info("... cwip: %d%% done..." % int(100*done / log_count_total))
                log_count = log_count_total / 20
            else:
                log_count -= 1

            if p.start is None: continue

            ## compute CWIP:
            p.CWIP=0
            p.wip_at_arrival=0
            for i in jobs:
                if i==p: continue
                if i.start is None: continue  ## i can be ignored (?)
                if i.arrival > p.start: break ## i and following can be ignored

                xip = max(0,
                          (min(i.completion,p.start)-max(i.start,p.arrival))/(i.completion-i.start)
                         )
                p.CWIP += xip * i.workload

                if i.arrival > p.arrival: continue                     ## not arrived yet
                if i.completion < p.arrival: continue                  ## already gone
                if i.start >= p.arrival: p.wip_at_arrival += i.workload## here, not started
                else: ## started
                    xip = max(0,1 - (p.arrival - i.start)/i.workload)  ## what should be left
                    p.wip_at_arrival += xip * i.workload

            ## compute WCap:
            if p.EWT() != 0: p.WCap = p.CWIP / p.EWT()
    
# ============================================================================
# Class for a Machine
# ============================================================================
class Machine:
    INFINITY=1e6
    
    def __init__(self,idf,speed=1,batch_size=1,
                     setup=lambda: 0,
                     recipes=[]):
        # init values:
        self.idf = idf              ## ID 
        self.speed = speed          ## speed-factor (2 means twice faster than normal)
        self.batch_size = batch_size## size of a batch
        self.setup = setup          ## function to define setup before handling a job
                                    ## (that may model manual handling or down-times)
        self.recipes = recipes      ## set of known recipes

        # operational values:
        self.cmax = 0 ## current completion time

    def __str__(self):
        return str(self.idf)

    ## -----------------------------------------------------------------------------
    def isCompatible(self, job):
        if job.recipe is None: return True
        return job.recipe in self.recipes

    ## -----------------------------------------------------------------------------
    def scheduleTo(self, t):
        self.cmax = t + self.setup()

    ## -----------------------------------------------------------------------------
    @staticmethod
    def read(filename):
        ''' Read machines from file as a ;-separated file.'''
        log.info('Reading machines from %s...' % filename)
        machines = []
        policy_name = None
        policy = None
        for line in fileinput.input(filename):
            try:
                line = line.strip()
                if line == "": continue ## skip blank lines
                if line[0] == "#": continue ## skip comments

                if line[0] == "@":
                    # score function:
                    _,policy_name = line.split(";")
                    policy = define_policy(policy_name)
                else:
                    # machine:
                    (idf,speed,batch_size,setup,recipes) = line.split(";")
                    if idf == 'idf': continue ## skip title line
                    m = Machine(idf, float(speed), int(batch_size), 
                                eval(setup), eval(recipes))
                    machines.append(m)
            except ValueError as e:
                log.error("%s:%d: %s" % (filename, fileinput.lineno(),e))

        log.info('... %d machines' % len(machines))
        log.info('... score: %s' % policy_name)
        return(machines,policy)


## -----------------------------------------------------------------------------
def define_policy(policy_name):
    policy = None
    if policy_name == "IDF":
        policy = lambda j,t: int(j.idf)
    elif policy_name == "FIFO":
        policy = lambda j,t: j.arrival
    elif policy_name == "PRIO":
        policy = lambda j,t: 1000*j.priority + j.arrival/1000
    elif policy_name == "WAIT-PRIO":
        policy = \
          lambda j,t: (j.arrival-t) * j.priority if (j.arrival > t) \
          else (j.arrival-t) / j.priority
    else:
        policy = eval(policy_name)
    return(policy)
    
## -----------------------------------------------------------------------------
def schedule(jobs, machines, policy=lambda j,t: -j.arrival):
    '''Schedule jobs on machines. Policy is non-idling whenever
    possible. Job-order is defined by policy function [t is current time,
    j is a job]'''
        
    log.info('Scheduling WIP ...')
    
    done = [] # jobs done
    log_count_total = len(jobs)
    log_count = 0
    while len(jobs) > 0:
        if log_count == 0:
            log.info("... scheduling: %d%% done..." % int(100*len(done) / log_count_total))
            log_count = log_count_total / 20
        else:
            log_count -= 1

        # next available machine:
        machines.sort(key=lambda m: m.cmax)
        m = machines[0]
        t = m.cmax ## current time
        if t >= Machine.INFINITY:
            log.warning("Cannot schedule some jobs.")
            break

        # sort job by score:        
        jobs.sort(key=lambda j: policy(j,t))
        job = None           ## current job to schedule
        job_idx = -1         ## its index

        # look for the best compatible job that is arrived before t:
        for i,j in enumerate(jobs):
            if j.arrival <= t and m.isCompatible(j):
                job_idx = i
                job = j
                break
        if job is None:
            ## nothing to do right now, look for the first available job
            t = Machine.INFINITY
            for i,j in enumerate(jobs):
                if j.arrival < t and m.isCompatible(j):
                    job_idx = i
                    job = j
                    t = j.arrival
        if job is None:
            # nothing compatible with m: "close" it
            m.scheduleTo(Machine.INFINITY)
        else:
            if m.batch_size > 1:
                # fill batch:
                batch = [job_idx]
                for i,j in enumerate(jobs[job_idx+1:],start=job_idx+1):
                    if j == job: continue
                    if j.batch_family == job.batch_family \
                      and j.arrival < t \
                      and m.isCompatible(j):
                        batch.append(i)
                        if len(batch) == m.batch_size: break ## batch is full

                # schedule batch:
                d = t + max([jobs[j].duration for j in batch])
                for i in batch: jobs[i].scheduleAt(t,completion=d,machine=m)
                m.scheduleTo(d)
                batch.reverse() ## must start with bigger index to not change smaller ones
                for i in batch:
                    done.append(jobs[i])
                    del jobs[i]

            else:
                # schedule single job at time t:
                d = t + job.duration/m.speed
                job.scheduleAt(t,completion=d,machine=m)
                m.scheduleTo(d)
                done.append(job)
                del jobs[job_idx]
            
    jobs.extend(done)
    return(jobs)
    
# ============================================================================
# Program setup
# ----------------------------------------------------------------------------

## log:
logging.basicConfig(format="%(asctime)s %(levelname)-8s %(message)s",
                        datefmt='%Y-%m-%d %H:%M:%S')
log = logging.getLogger("")
log.setLevel(logging.DEBUG)

if sys.version_info < (3, 0):
    log.fatal("python 3 ou supérieur requis")
    exit(1)

## arguments:
parser = argparse.ArgumentParser()
parser.add_argument("jobfile", action="store")
parser.add_argument("machinefile", action="store")
parser.add_argument("--policy", action="store") ## to force a policy
parser.add_argument("--nocwip", action="store") ## disable cwip if not None
args = parser.parse_args()    
log.info('Started!')

# ----------------------------------------------------------------------------
## processing:
jobs = Job.read(args.jobfile)
machines,policy = Machine.read(args.machinefile)
if args.policy is not None:
    log.info("Force policy: %s" % args.policy)
    policy = define_policy(args.policy)
jobs = schedule(jobs,machines, policy=policy)
if args.nocwip is None: Job.cwip(jobs)
Job.write(jobs)

# ----------------------------------------------------------------------------
# Done
# ----------------------------------------------------------------------------
log.info('Done!')
exit(0)

